<?php

namespace Drupal\cas_server\Form;

use Drupal\cas_server\Configuration\ConfigHelper;
use Drupal\cas_server\EventSubscriber\CASResponseCookies;
use Drupal\cas_server\Logger\DebugLogger;
use Drupal\cas_server\Ticket\TicketFactory;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\DependencyInjection\AutowireTrait;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\user\UserAuthInterface;

/**
 * Provides the CAS user login form.
 */
class UserLogin extends FormBase {

  // Remove need to static::create() method.
  use AutowireTrait;

  /**
   * Constructs a \Drupal\cas_server\Form\UserLogin object.
   *
   * @param \Drupal\cas_server\EventSubscriber\CASResponseCookies $responseCookies
   *   The returned response cookies manager.
   * @param \Drupal\cas_server\Configuration\ConfigHelper $configHelper
   *   The configuration helper.
   * @param \Drupal\cas_server\Logger\DebugLogger $logger
   *   The logger.
   * @param \Drupal\cas_server\Ticket\TicketFactory $ticketFactory
   *   The ticket factory.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity manager.
   * @param \Drupal\user\UserAuthInterface $authService
   *   The authentication provider.
   */
  public function __construct(
    protected CASResponseCookies $responseCookies,
    protected ConfigHelper $configHelper,
    protected DebugLogger $logger,
    protected TicketFactory $ticketFactory,
    protected TimeInterface $time,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected UserAuthInterface $authService
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cas_server_user_login';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $service = '') {

    // Identifying credential field. Can be configured to access username,
    // email address, or both.
    $form['username'] = [
      '#type' => 'textfield',
      '#title' => $this->configHelper->getAuthenticationSourceFieldTitle(),
      '#size' => 60,
      '#maxlength' => 128,
      '#required' => TRUE,
      '#attributes' => [
        'autocomplete' => 'username',
      ],
    ];

    $form['password'] = [
      '#type' => 'password',
      '#title' => $this->t('Password'),
      '#size' => 60,
      '#required' => TRUE,
      '#attributes' => [
        'autocomplete' => 'current-password',
      ],
    ];

    // Only regenerate lt if we're here on an initial build, not a submission.
    $request = $this->getRequest();

    // A login ticket is a string that is provided by /login as a credential
    // requestor and passed to /login as a credential acceptor for
    // username/password authentication. Its purpose is to prevent the
    // replaying of credentials due to bugs in web browsers. Once used, a LT
    // must not be allowed to be used again.
    // https://apereo.github.io/cas/6.6.x/protocol/CAS-Protocol-V2-Specification.html#351-login-ticket-properties
    if ($request->getMethod() != 'POST') {
      // @todo LT must be generated and saved, just like any other so that it
      // can be tracked as per above usage description.
      $lt = 'LT-' . Crypt::randomBytesBase64(32);
      $_SESSION['cas_lt'] = $lt;
    }

    // @todo Provide LT via ajax or other means so that the form can be cached.
    $form['lt'] = [
      '#type' => 'hidden',
      '#value' => $_SESSION['cas_lt'] ?? '',
    ];

    $form['service'] = [
      '#type' => 'hidden',
      '#value' => $service,
    ];

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    if ($this->configHelper->getShowResetPassword()) {
      $form['request_password'] = [
        '#theme' => 'item_list',
        '#items' => [
          [
            '#type' => 'link',
            '#title' => $this->t('Reset your password'),
            '#url' => Url::fromRoute('user.pass', [], [
              'attributes' => [
                'title' => $this->t('Send password reset instructions via email.'),
                'class' => ['request-password-link'],
              ],
            ]),
          ],
        ],
      ];
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $username = trim($form_state->getValue('username'));
    if ($this->configHelper->getAuthenticationSourceField() != 'name') {
      $users = $this->entityTypeManager->getStorage('user')
        ->loadByProperties(['mail' => $username]);
      if ($users) {
        $user = reset($users);
        $username = $user->getAccountName();

      }
      elseif ($this->configHelper->getAuthenticationSourceField() == 'mail') {
        $form_state->setErrorByName(
          'username',
          $this->t('Invalid username or password.')
        );
        return;
      }
    }
    $password = trim($form_state->getValue('password'));
    if (!($uid = $this->authService->authenticate($username, $password))) {
      $form_state->setErrorByName(
        'username',
        $this->t('Invalid username or password.')
      );
      return;
    }

    // Get the service entity. Service should be validated before use.
    $service = $form_state->getValue('service');
    if (!empty($service)) {
      // Get and validate the service.
      $serviceEntity = $this->configHelper->loadServiceFromUri($service);
      if (!$serviceEntity) {
        $form_state->setErrorByName(
          'service',
          $this->t('Service provided is invalid for authentication here.')
        );
        return;
      }

      // Check that a user with uid is permitted to authenticate with service.
      $account = $this->entityTypeManager->getStorage('user')->load($uid);
      if (!$serviceEntity->accountPermitted($account)) {
        $form_state->setErrorByName(
          'username',
          $this->t('Access denied: Your account does not have the required permissions to log in to this service.')
        );
        return;
      }
    }

    // Pass validated values onto submit.
    $form_state->setValue('uid', $uid);
    $form_state->setValue('username', $username);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    if ((empty($_SESSION['cas_lt'])) || $form_state->getValue('lt') != $_SESSION['cas_lt']) {
      $this->messenger()->addError(
        $this->t('Login ticket invalid. Refresh page and try again.')
      );
      $form_state->setRedirectUrl(Url::fromRoute('cas_server.login'));

      return;
    }

    $service = $form_state->getValue('service');
    if ($uid = $form_state->getValue('uid')) {
      $account = $this->entityTypeManager->getStorage('user')->load($uid);
      user_login_finalize($account);

      // Add TGC if configured to do so.
      if (empty($service) || $this->configHelper->verifyServiceForSso($service)) {
        if ($this->configHelper->shouldUseTicketGrantingTicket()) {
          $tgt = $this->ticketFactory->createTicketGrantingTicket();
          $this->responseCookies->setCookie(
            'cas_tgc',
            $tgt->getId(),
            $this->configHelper
              ->getTicketGrantingTicketTimeout($this->time->getRequestTime()),
            Url::fromUri('internal:/cas')->toString()
          );
        }
      }

      // Redirect user to requested service.
      if (!empty($service)) {
        $st = $this->ticketFactory->createServiceTicket($service, TRUE);
        $url = Url::fromRoute('cas_server.login', [], [
          'query' => [
            'service' => $service,
            'ticket' => $st->getId(),
          ],
        ]);
        $form_state->setRedirectUrl($url);

        return;
      }

      // No service to redirect to, show them confirmed login.
      $form_state->setRedirectUrl(Url::fromRoute('cas_server.login'));
      return;
    }

    // User has failed to log in.
    $this->messenger()->addError(
      $this->t('Bad username/password combination given.')
    );
    $form_state->setRedirectUrl(Url::fromRoute('cas_server.login'));
  }

}
